package dev.pfaff.log4truth.compat.slf4j;

import dev.pfaff.log4truth.NamedLogger;
import dev.pfaff.log4truth.StringSupplier;
import org.slf4j.Logger;
import org.slf4j.Marker;

import java.lang.ref.SoftReference;
import java.util.*;
import java.util.function.IntFunction;

public record SLF4JToL4TLogger(NamedLogger logger) implements Logger {
	public SLF4JToL4TLogger(String name) {
		this(new NamedLogger(name));
	}

	@Override
	public String getName() {
		return logger.name();
	}

	@Override
	public boolean isTraceEnabled() {
		return true;
	}

	@Override
	public boolean isTraceEnabled(Marker marker) {
		return true;
	}

	@Override
	public boolean isDebugEnabled() {
		return true;
	}

	@Override
	public boolean isDebugEnabled(Marker marker) {
		return true;
	}

	@Override
	public boolean isInfoEnabled() {
		return true;
	}

	@Override
	public boolean isInfoEnabled(Marker marker) {
		return true;
	}

	@Override
	public boolean isWarnEnabled() {
		return true;
	}

	@Override
	public boolean isWarnEnabled(Marker marker) {
		return true;
	}

	@Override
	public boolean isErrorEnabled() {
		return true;
	}

	@Override
	public boolean isErrorEnabled(Marker marker) {
		return true;
	}

	private void log(List<String> level, String msg) {
		logger.log(StringSupplier.of(msg), level);
	}

	// TODO: maybe use StringConcatFactory?
	record Template(List<String> fragments) {}

	private static final WeakHashMap<String, SoftReference<Template>> TEMPLATE_CACHE = new WeakHashMap<>();

	private static SoftReference<Template> processTemplate(String format) {
		var fragments = new ArrayList<String>();
		int start = 0;
		int cursor = 0;
		boolean escaped = false;
		while (cursor < format.length()-1) {
			var c = format.charAt(cursor++);
			if (c == '\\') {
				escaped = !escaped;
			} else if (!escaped && c == '{') {
				c = format.charAt(cursor);
				if (c == '}') {
					cursor++;
					fragments.add(format.substring(start, cursor-2));
					start = cursor;
				}
			}
		}
		fragments.add(format.substring(start));
		return new SoftReference<>(new Template(List.copyOf(fragments)));
	}

	private void logFormat(List<String> level, String format, int argCount, IntFunction<Object> args) {
		var template = TEMPLATE_CACHE.computeIfAbsent(format, SLF4JToL4TLogger::processTemplate).get();
		if (template == null) {
			var templateRef = processTemplate(format);
			TEMPLATE_CACHE.replace(format, templateRef);
			template = Objects.requireNonNull(templateRef.get());
		}
		var template_ = template;
		Throwable throwable;
		if (template_.fragments.size() - 1 < argCount && args.apply(argCount-1) instanceof Throwable t) {
			throwable = t;
		} else {
			throwable = null;
		}
		logger.log(() -> {
			var sb = new StringBuilder(format.length());
			var fragments = template_.fragments;
			int i = 0;
			for (; i < fragments.size() - 1; i++) {
				sb.append(fragments.get(i));
				if (i < argCount) {
					sb.append(args.apply(i));
				} else {
					sb.append("{}");
				}
			}
			sb.append(fragments.get(fragments.size() - 1));
			if (i < argCount) {
				int upr = throwable != null ? argCount-1 : argCount;
				for (; i < upr; i++) {
					sb.append("<extra arg>").append(args.apply(i));
				}
			}
			return sb.toString();
		}, level);
	}

	private void log(List<String> level, String format, Object arg) {
		logFormat(level, format, 1, i -> arg);
	}

	private void log(List<String> level, String format, Object arg1, Object arg2) {
		logFormat(level, format, 2, i -> i == 0 ? arg1 : arg2);
	}

	private void log(List<String> level, String format, Object... args) {
		logFormat(level, format, args.length, i -> args[i]);
	}

	private void log(List<String> level, String msg, Throwable throwable) {
		logger.log(StringSupplier.of(msg), throwable, level);
	}

	@Override
	public void trace(String msg) {
		log(List.of("TRACE"), msg);
	}

	@Override
	public void trace(String format, Object arg) {
		log(List.of("TRACE"), format, arg);
	}

	@Override
	public void trace(String format, Object arg1, Object arg2) {
		log(List.of("TRACE"), format, arg1, arg2);
	}

	@Override
	public void trace(String format, Object... args) {
		log(List.of("TRACE"), format, args);
	}

	@Override
	public void trace(String msg, Throwable t) {
		log(List.of("TRACE"), msg, t);
	}

	@Override
	public void trace(Marker marker, String msg) {
		log(List.of("TRACE", marker.getName()), msg);
	}

	@Override
	public void trace(Marker marker, String format, Object arg) {
		log(List.of("TRACE", marker.getName()), format, arg);
	}

	@Override
	public void trace(Marker marker, String format, Object arg1, Object arg2) {
		log(List.of("TRACE", marker.getName()), format, arg1, arg2);
	}

	@Override
	public void trace(Marker marker, String format, Object... args) {
		log(List.of("TRACE", marker.getName()), format, args);
	}

	@Override
	public void trace(Marker marker, String msg, Throwable t) {
		log(List.of("TRACE", marker.getName()), msg, t);
	}

	@Override
	public void debug(String msg) {
		log(List.of("DEBUG"), msg);
	}

	@Override
	public void debug(String format, Object arg) {
		log(List.of("DEBUG"), format, arg);
	}

	@Override
	public void debug(String format, Object arg1, Object arg2) {
		log(List.of("DEBUG"), format, arg1, arg2);
	}

	@Override
	public void debug(String format, Object... args) {
		log(List.of("DEBUG"), format, args);
	}

	@Override
	public void debug(String msg, Throwable t) {
		log(List.of("DEBUG"), msg, t);
	}

	@Override
	public void debug(Marker marker, String msg) {
		log(List.of("DEBUG", marker.getName()), msg);
	}

	@Override
	public void debug(Marker marker, String format, Object arg) {
		log(List.of("DEBUG", marker.getName()), format, arg);
	}

	@Override
	public void debug(Marker marker, String format, Object arg1, Object arg2) {
		log(List.of("DEBUG", marker.getName()), format, arg1, arg2);
	}

	@Override
	public void debug(Marker marker, String format, Object... args) {
		log(List.of("DEBUG", marker.getName()), format, args);
	}

	@Override
	public void debug(Marker marker, String msg, Throwable t) {
		log(List.of("DEBUG", marker.getName()), msg, t);
	}

	@Override
	public void info(String msg) {
		log(List.of("INFO"), msg);
	}

	@Override
	public void info(String format, Object arg) {
		log(List.of("INFO"), format, arg);
	}

	@Override
	public void info(String format, Object arg1, Object arg2) {
		log(List.of("INFO"), format, arg1, arg2);
	}

	@Override
	public void info(String format, Object... args) {
		log(List.of("INFO"), format, args);
	}

	@Override
	public void info(String msg, Throwable t) {
		log(List.of("INFO"), msg, t);
	}

	@Override
	public void info(Marker marker, String msg) {
		log(List.of("INFO", marker.getName()), msg);
	}

	@Override
	public void info(Marker marker, String format, Object arg) {
		log(List.of("INFO", marker.getName()), format, arg);
	}

	@Override
	public void info(Marker marker, String format, Object arg1, Object arg2) {
		log(List.of("INFO", marker.getName()), format, arg1, arg2);
	}

	@Override
	public void info(Marker marker, String format, Object... args) {
		log(List.of("INFO", marker.getName()), format, args);
	}

	@Override
	public void info(Marker marker, String msg, Throwable t) {
		log(List.of("INFO", marker.getName()), msg, t);
	}

	@Override
	public void warn(String msg) {
		log(List.of("WARN"), msg);
	}

	@Override
	public void warn(String format, Object arg) {
		log(List.of("WARN"), format, arg);
	}

	@Override
	public void warn(String format, Object arg1, Object arg2) {
		log(List.of("WARN"), format, arg1, arg2);
	}

	@Override
	public void warn(String format, Object... args) {
		log(List.of("WARN"), format, args);
	}

	@Override
	public void warn(String msg, Throwable t) {
		log(List.of("WARN"), msg, t);
	}

	@Override
	public void warn(Marker marker, String msg) {
		log(List.of("WARN", marker.getName()), msg);
	}

	@Override
	public void warn(Marker marker, String format, Object arg) {
		log(List.of("WARN", marker.getName()), format, arg);
	}

	@Override
	public void warn(Marker marker, String format, Object arg1, Object arg2) {
		log(List.of("WARN", marker.getName()), format, arg1, arg2);
	}

	@Override
	public void warn(Marker marker, String format, Object... args) {
		log(List.of("WARN", marker.getName()), format, args);
	}

	@Override
	public void warn(Marker marker, String msg, Throwable t) {
		log(List.of("WARN", marker.getName()), msg, t);
	}

	@Override
	public void error(String msg) {
		log(List.of("ERROR"), msg);
	}

	@Override
	public void error(String format, Object arg) {
		log(List.of("ERROR"), format, arg);
	}

	@Override
	public void error(String format, Object arg1, Object arg2) {
		log(List.of("ERROR"), format, arg1, arg2);
	}

	@Override
	public void error(String format, Object... args) {
		log(List.of("ERROR"), format, args);
	}

	@Override
	public void error(String msg, Throwable t) {
		log(List.of("ERROR"), msg, t);
	}

	@Override
	public void error(Marker marker, String msg) {
		log(List.of("ERROR", marker.getName()), msg);
	}

	@Override
	public void error(Marker marker, String format, Object arg) {
		log(List.of("ERROR", marker.getName()), format, arg);
	}

	@Override
	public void error(Marker marker, String format, Object arg1, Object arg2) {
		log(List.of("ERROR", marker.getName()), format, arg1, arg2);
	}

	@Override
	public void error(Marker marker, String format, Object... args) {
		log(List.of("ERROR", marker.getName()), format, args);
	}

	@Override
	public void error(Marker marker, String msg, Throwable t) {
		log(List.of("ERROR", marker.getName()), msg, t);
	}
}
